Require Import computable_total_ordering.

Module Functional_maps(Domain : ORDERED_SET).
Import Domain.

Definition domain := Domain.t.
Notation domain_le := Domain.le.
Notation domain_le_refl := Domain.le_refl.
Notation domain_le_trans := Domain.le_trans.
Notation compare := Domain.compare.

Section Functional_maps_codomain.
Variable codomain : Set.

Definition func_add
    (key : domain) (img : codomain)
    (f : domain -> (option codomain))  (lookup : domain) :=
    match (compare lookup key) with
    | EQUAL _ _ => (Some img)
    | _ => f lookup
    end.
Definition func_remove (key : domain)
     (f : (domain -> (option codomain))) (lookup : domain) :
     (option codomain) :=
    match (compare key lookup) with
    | EQUAL _ _ => None
    | _ => (f lookup)
    end.
End Functional_maps_codomain.
End Functional_maps.

Module Type ABSTRACT_MAP.
Declare Module Domain : ORDERED_SET.
Notation domain := Domain.t.
Notation compare := Domain.compare.

Parameter t : Set -> Set.

Section ABSTRACT_MAP_codomain.
Variable codomain : Set.

Parameter empty : unit -> (t codomain).
Parameter find : domain -> (t codomain) -> (option codomain).

Parameter add : domain -> codomain -> (t codomain) -> (t codomain).
Definition func_add
    (key : domain) (img : codomain)
     (f : domain -> (option codomain)) (lookup : domain) :=
    match (compare lookup key) with
    | EQUAL _ _ => (Some img)
    | _ => f lookup
    end.
Axiom add_correct :
    forall key : domain, forall img : codomain,
    forall l : (t codomain),
    forall lookup : domain,
    (find lookup (add key img l)) =
    (func_add key img (fun lookup1 => find lookup1 l) lookup).

Parameter remove : domain -> (t codomain) -> (t codomain).
Definition func_remove (key : domain)
    (f : (domain -> (option codomain))) (lookup : domain): (option codomain):=
    match (compare key lookup) with
    | EQUAL _ _ => None
    | _ => (f lookup)
    end.
Axiom remove_correct :
    forall key : domain, forall m : (t codomain),
    forall lookup : domain,
    (func_remove key (fun lookup1 => find lookup1 m) lookup) =
    (find lookup (remove key m)).
End ABSTRACT_MAP_codomain.

Section ABSTRACT_MAP_mapi1_opt.
Variable codomain1 codomain2 : Set.

Parameter mapi1_opt : (domain -> codomain1 -> (option codomain2)) ->
  (t codomain1) -> (t codomain2).

Variable f : domain -> codomain1 -> (option codomain2).

Definition func_mapi1_opt
    (mapf : domain -> (option codomain1))
    (lookup : domain) : (option codomain2) :=
    match mapf lookup with
    | None => None
    | Some x => f lookup x
    end.

Axiom mapi1_opt_correct :
    (forall d1 d2 : domain,
    Domain.le d1 d2 -> Domain.le d2 d1 -> 
    forall c : codomain1,
    (f d1 c) = (f d2 c)) ->
    forall m : (t codomain1),
    forall lookup : domain,
    (find codomain2 lookup (mapi1_opt f m)) =
    (func_mapi1_opt (fun lookup1 => find codomain1 lookup1 m) lookup).
End ABSTRACT_MAP_mapi1_opt.

Section ABSTRACT_MAP_mapi2_opt.

Variable codomain1 codomain2 codomain3 : Set.

Definition func_mapi2_opt
    (f1 : domain -> codomain1 -> (option codomain3))
    (f2 : domain -> codomain2 -> (option codomain3))
    (g : domain -> codomain1 -> codomain2 -> (option codomain3))

    (m1 : domain -> (option codomain1))
    (m2 : domain -> (option codomain2))
    (lookup : domain) : (option codomain3) :=
    match (m1 lookup) with
    | (Some c1) =>
      match (m2 lookup) with
      | (Some c2) => (g lookup c1 c2)
      | None => (f1 lookup c1)
      end
    | None =>
      match (m2 lookup) with
      | (Some c2) => (f2 lookup c2)
      | None => None
      end
    end.

Parameter mapi2_opt :
    (domain -> codomain1 -> (option codomain3)) ->
    (domain -> codomain2 -> (option codomain3)) ->
    (domain -> codomain1 -> codomain2 -> (option codomain3)) ->
    (t codomain1) -> (t codomain2) -> (t codomain3).

Axiom mapi2_opt_correct:
    forall f1 : domain -> codomain1 -> (option codomain3),
    forall f2 : domain -> codomain2 -> (option codomain3),
    forall g : domain -> codomain1 -> codomain2 -> (option codomain3),

    (forall d1 d2 : domain,
    (Domain.le d1 d2) -> (Domain.le d2 d1) ->
    forall c : codomain1,
    (f1 d1 c) = (f1 d2 c)) ->

    (forall d1 d2 : domain,
    (Domain.le d1 d2) -> (Domain.le d2 d1) ->
    forall c : codomain2,
    (f2 d1 c) = (f2 d2 c)) ->

    (forall d1 d2 : domain,
    (Domain.le d1 d2) -> (Domain.le d2 d1) ->
    forall c1 : codomain1,
    forall c2 : codomain2,
    (g d1 c1 c2) = (g d2 c1 c2)) ->

    forall m1 : (t codomain1),
    forall m2 : (t codomain2),
    forall lookup : domain,
    (find codomain3 lookup (mapi2_opt f1 f2 g m1 m2)) =
    (func_mapi2_opt f1 f2 g
      (fun lookup1 => (find codomain1 lookup1 m1))
      (fun lookup1 => (find codomain2 lookup1 m2)) lookup).
End ABSTRACT_MAP_mapi2_opt.

(*
Parameter mapi2_opt :
  forall codomain1 codomain2 codomain3 : Set,
  (domain -> codomain1 -> (option codomain3)) ->
  (domain -> codomain2 -> (option codomain3)) ->
  (domain -> codomain1 -> codomain2 -> (option codomain3)) ->
  (t codomain1) -> (t codomain2) -> (t codomain3).*)
End ABSTRACT_MAP.
